FTP.Client = {}
FTP.Queue = {}

function FTP.Login(Name, Password)
	local User = FTP.Users[Name]
	if User then
		if User[1] == Password then
			local Path = User[2]
			while Path:sub(-1, -1) == "/" do
				Path = Path:sub(1, -2)
			end
			return Path
		end
	end
end

function FTP.Print(...)
	local Arguments = {}
	for _, Argument in pairs({...}) do
		table.insert(Arguments, tostring(Argument))
	end
	print("[FTP] "..table.concat(Arguments, "	"))
end

function FTP.InitServer()
	local Error
	FTP.Server, Error = socket.bind("*", FTP.Configuration.Port)

	if FTP.Server then
		FTP.Server:setoption("tcp-nodelay", true)
		FTP.Server:settimeout(0)

		local IP, Port = FTP.Server:getsockname()
		FTP.Print("Initialized FTP server  "..IP..":"..Port)
	else
		FTP.Print("Failed to start server: "..Error)
	end
end

function FTP.Start()
	if not FTP.Server then
		FTP.Client = {}
		FTP.Queue = {}
		FTP.InitServer("127.0.0.1", 21, 256)
	end
end

function FTP.Stop()
	if FTP.Server then
		for _, Client in pairs(FTP.Client) do
			Client:Close()
		end
		FTP.Server:close()
		FTP.Server = nil
		FTP.Print("Closed FTP Server")
	end
end

function FTP.RemoveQueue(id, success)
	local Queue = FTP.Queue[id]
	if Queue then
		Queue.Socket:close()
		Queue.file:close()
		Queue = nil
	end
end

local Client = {}
local ClientMetatable = {__index = Client}

function FTP.UpdateServer()
	if not FTP.Server then
		return nil
	end

	local Socket = FTP.Server:accept()
	if Socket then
		local IP, Port = Socket:getpeername()
		FTP.Print("New Client "..IP..":"..Port)

		Socket:settimeout(0)
		Socket:send("220 Hello, this is Starkkz's FTP Service\n")

		local Client = setmetatable({
			Socket = Socket,
			RenameFrom = "",
			User = "",
			Password = "",
			Dir = "",
			CONMODE = 0,
			DIP = "",
			DPORT = 0,
			CMDQUEUE = {},
			ID = #FTP.Client + 1,
		}, ClientMetatable)
		FTP.Client[Client.ID] = Client
	end
end

function FTP.UpdateClients()
	for id, Client in pairs(FTP.Client) do
		local Line, Error = Client.Socket:receive("*l")
		if Error == "closed" then
			local IP, Port = Client.Socket:getpeername()
			Client:Close()
			FTP.Client[id] = nil
			FTP.Print("Client "..IP..":"..Port.." disconnected")
		elseif Line then
			table.insert(Client.CMDQUEUE, Line)
		end

		for CMDID, CMD in pairs(Client.CMDQUEUE) do
			if Client:Parse(CMD) then
				Client.CMDQUEUE[CMDID] = nil
			end
		end
	end
end

function Client:ip()
	local ip, port = self.Socket:getsockname()
	return ip
end

function Client:port()
	local ip, port = self.Socket:getsockname()
	return port
end

function Client:DataIP()
	if self.DataSocket then
		local ip, port = self.DataSocket:getsockname()
		return ip
	end
end

function Client:DataPort()
	if self.DataSocket then
		local ip, port = self.DataSocket:getsockname()
		return port
	end
end

function Client:ConnectData()
	if self.DataSocket then
		return true
	end

	if self.CONMODE == 1 then
		if self.PasvServer then
			local Sock = self.PasvServer:accept()
			if Sock then
				Sock:settimeout(0, "b")
				Sock:settimeout(0, "t")
				self.DataSocket = Sock
				return true
			end
		end
	elseif self.CONMODE == 2 then
		if #self.DIP > 0 and self.DPORT > 0 then
			local Sock, e = socket.tcp()
			if Sock then
				Sock:settimeout(2000, "b")
				Sock:settimeout(2000, "t")

				local Success, Error = Sock:connect(self.DIP, self.DPORT)
				if Success then
					Sock:settimeout(0, "b")
					Sock:settimeout(0, "t")
					self.DataSocket = Sock
					return true
				end
				Sock:close()
			end
		end
	end
end

function Client:CloseData()
	if self.DataSocket then
		self.DataSocket:close()
		self.DataSocket = nil
	end
end

function Client:ClosePasv()
	if self.PasvServer then
		self:CloseData()
		self.PasvServer:close()
		self.PasvServer = nil
	end
end

function Client:CloseConMode()
	if Client.CONMODE == 1 then
		self:ClosePasv()
	elseif Client.CONMODE == 2 then
		self:CloseData()
	end
end

function Client:Close()
	self:CloseData()
	self:ClosePasv()
	self.Socket:close()
end

function Client:send(Datagram)
	local R, W, E = socket.select(nil, {self.Socket}, 5000)
	if not E then
		local Index, Next = next(W)
		if Next == self.Socket then
			self.Socket:send(Datagram)
		end
	else
		FTP.Print("Send error: "..E)
	end
end

function Client:Parse(Line)
	if not Line then
		return nil
	end

	if string.sub(Line, 1, 4) == "USER" then
		local UserName = string.sub(Line, 6)
		if UserName:len() > 0 then
			self.User = UserName
			self:send("331 Password required to access user account "..UserName..".\n")
		end
		return true
	elseif string.sub(Line, 1, 4) == "PASS" then
		local Password = string.sub(Line, 6)
		if Password:len() > 0 then
			local Path = FTP.Login(self.User, Password)
			if Path then
				self.Password = Password
				self.Logged = true
				self.Dir = Path
				self.DefaultDir = Path
				self:send("230 Logged in.\n")
			else
				self:send("530 Login authentication failed\n")
			end
		else
			self:send("331 Password required to access user account "..UserName..".\n")
		end
		return true
	elseif string.sub(Line, 1, 3) == "CWD" then
		if self.Logged then
			local Folder = Line:sub(5)
			while Folder:sub(1, 1) == "/" do
				Folder = Folder:sub(2)
			end

			if Folder == ".." then
				local find
				for i = 1, #self.Dir do
					if string.sub(self.Dir, i, i) == "/" then
						find = i
					end
				end

				if find then
					self.Dir = string.sub(self.Dir, 1, find - 1) or self.DefaultDir
					self:send('250 "/'..self.Dir:sub(#self.DefaultDir + 1)..'" is new working directory\n')
				else
					self.Dir = self.DefaultDir
					self:send('250 "/'..self.Dir:sub(#self.DefaultDir + 1)..'" is new working directory\n')
				end
			elseif #Folder > 0 then
				if Line:sub(5, 5) == "/" then
					self.Dir = self.DefaultDir .."/".. Folder
				elseif #Folder <= 0 then
					self.Dir = self.DefaultDir
				else
					self.Dir = self.Dir.."/"..Folder
				end
				self:send('250 "/'..self.Dir:sub(#self.DefaultDir + 1)..'" is new working directory\n')
			else
				self.Dir = self.DefaultDir
				self:send('250 "/'..self.Dir:sub(#self.DefaultDir + 1)..'" is new working directory\n')
			end
		else
			self:send("530 You aren't logged in\n")
		end
		return true
	elseif string.sub(Line, 1, 4) == "SYST" then
		self:send("215 Lua Virtual Machine\n")
		return true
	elseif string.sub(Line, 1, 4) == "TYPE" then
		if self.Logged then
			local NewType = string.sub(Line, 6)
			if NewType == "I" then
				self:send("200 TYPE is now 8-bit binary\n")
			elseif NewType == "A" then
				self:send("200 TYPE is now ASCII\n")
			end
		else
			self:send("530 You aren't logged in\n")
		end
		return true
	elseif string.sub(Line, 1, 4) == "LIST" then
		if self.Logged then
			if not self.LSTART then
				self.LSTART = socket.gettime()
			end

			if self:ConnectData() then
				self:send("150 Opening ASCII mode data connection for file list.\n")
				local Dir
				if self.Dir:len() > 0 then
					Dir = self.Dir
				else
					Dir = self.DefaultDir
				end

				if FTP.IsDir(Dir) then
					for File in FTP.EnumDir(Dir) do
						if File ~= "." and File ~= ".." then
							if FTP.IsDir(Dir.."/"..File) then
								self.DataSocket:send("drwxr--r--	1	"..self.User.."	PC	"..self.ID.."	0	"..File.."\n")
							else
								local FileObject
								if #Dir > 0 then
									FileObject = FTP.Open(Dir.."/"..File, "rb")
								else
									FileObject = FTP.Open(File, "rb")
								end

								if FileObject then
									local Size = FileObject:seek("end"); FileObject:close()
									self.DataSocket:send("-rwxr--r--	1	"..self.User.."	PC	"..self.ID.."	"..Size.."	"..File.."\n")
								end
							end
						end
					end
				end
				self:send("226 Transfer complete.\n")
				self:CloseData()
				self.LSTART = nil
				return true
			elseif socket.gettime() - self.LSTART >= 2 then
				self:send("425 Can't open data connection\n")
				self.LSTART = nil
				return true
			end
		else
			self:send("530 You aren't logged in\n")
			return true
		end
	elseif string.sub(Line, 1, 4) == "PASV" then
		if self.Logged then
			self:ClosePasv()
			local sock, e = socket.tcp()
			if sock then
				local success, e = sock:bind("*", 0)
				if success then
					success, e = sock:listen()
					if success then
						sock:settimeout(0, "b")
						sock:settimeout(0, "t")

						local ip, port = sock:getsockname()
						local ports = {}
						ports[1] = math.floor(port/256)
						ports[2] = port - ports[1]*256

						local ips = {}
						for n in FTP.Configuration.IP:gmatch("([^%p]+)") do
							table.insert(ips, n)
						end

						self:send("227 Entering Passive Mode ("..ips[1]..","..ips[2]..","..ips[3]..","..ips[4]..","..ports[1]..","..ports[2]..")\n")
						self.PasvServer = sock
						self.CONMODE = 1
					else
						sock:close()
					end
				else
					sock:close()
				end
			else
				sock:close()
			end
		else
			self:send("530 You aren't logged in\n")
		end
		return true
	elseif string.sub(Line, 1, 4) == "PORT" then
		if self.Logged then
			self:CloseConMode()
			local adr = {string.match(Line, "PORT (%d+),(%d+),(%d+),(%d+),(%d+),(%d+)")}
			if #adr >= 6 then
				local ip = adr[1].."."..adr[2].."."..adr[3].."."..adr[4]
				local port = tonumber(adr[5])*256 + tonumber(adr[6])

				if ip == self:ip() then
					self.DataSocket = DataSocket
					self:send("200 PORT command successful\n")
					self.DIP = ip
					self.DPORT = port
					self.CONMODE = 2
				else
					self:send("500 I won't open a connection to "..ip.." (only to "..self:ip()..")\n")
				end
			else
				self:send("500 I won't open a connection\n")
			end
		else
			self:send("530 You aren't logged in\n")
		end
		return true
	elseif string.sub(Line, 1, 3) == "MKD" then
		if self.Logged then
			local Folder = string.sub(Line, 5)
			local Dir
			while string.sub(Folder, 1, 1) == "/" do
				Folder = string.sub(Folder, 2)
			end

			if self.Dir:len() > 0 then
				Dir = self.Dir.."/"..Folder
			else
				Dir = Folder
			end

			local success, e = FTP.MkDir(Dir)
			if success then
				self:send('257 "'..Folder..'" : The directory was successfully created\n')
			else
				local e = e or "File exists"
				self:send("550 Can't create directory: "..e.."\n")
			end
		else
			self:send("530 You aren't logged in\n")
		end
		return true
	elseif string.sub(Line, 1, 3) == "RMD" then
		if self.Logged then
			local Folder = string.sub(Line, 5)
			local Dir
			while string.sub(Folder, 1, 1) == "/" do
				Folder = string.sub(Folder, 2)
			end

			if self.Dir:len() > 0 then
				Dir = self.Dir.."/"..Folder
			else
				Dir = Folder
			end

			local success, e = FTP.Remove(Dir)
			if success then
				self:send("250 The directory was successfully removed\n")
			else
				self:send("550 Can't remove directory: "..e.."\n")
			end
		else
			self:send("530 You aren't logged in\n")
		end
		return true
	elseif string.sub(Line, 1, 4) == "PWD" then
		if self.Logged then
			self:send('257 "/'..self.Dir:sub(#self.DefaultDir + 1)..'" is the current working directory\n')
		else
			self:send("530 You aren't logged in\n")
		end
		return true
	elseif string.sub(Line, 1, 4) == "RNFR" then
		local File = string.sub(Line, 6)
		local Dir
		if self.Dir:len() > 0 then
			Dir = self.Dir.."/"..File
		else
			Dir = File
		end

		if File:len() <= 0 then
			self:send("550 No file name\n")
		elseif File:len() > 0 then
			local FileObject = FTP.Open(Dir, "rb")
			if FileObject then
				FileObject:close()

				self:send("350 RNFR accepted - file exists, ready for destination\n")
				self.RenameFrom = File
			else
				self:send("550 Sorry, but that file doesn't exist\n")
			end
		end
		return true
	elseif string.sub(Line, 1, 4) == "RNTO" then
		if not self.RenameFrom or self.RenameFrom:len() <= 0 then
			self:send("503 Need RNFR before RNTO\n")
			return true
		end

		local File = string.sub(Line, 6)
		local RenameDir, SourceDir
		if self.Dir:len() > 0 then
			RenameDir = self.Dir.."/"..File
			SourceDir = self.Dir.."/"..self.RenameFrom
		else
			RenameDir = File
			SourceDir = self.RenameFrom
		end

		if File:len() <= 0 then
			self:send("550 No file name\n")
		elseif File:len() > 0 then
			local FileObject = FTP.Open(RenameDir, "rb")
			if FileObject then
				FileObject:close()

				self:send("550 RENAME Failed - destination file already exists\n")
			else
				self:send("250 File successfully renamed\n")
				os.rename(SourceDir, RenameDir)
			end
		end
		return true
	elseif string.sub(Line, 1, 4) == "DELE" then
		local File = string.sub(Line, 6)
		local Dir
		while string.sub(File, 1, 1) == "/" do
			File = string.sub(File, 2)
		end

		if File == "cs2dftp" then
			if self.Dir == "/sys/lua" then
				self:send("550 Could not delete FILENAME: blocked\n")
				return true
			end
		end

		if self.Dir:len() > 0 then
			Dir = self.Dir.."/"..File
		else
			Dir = File
		end

		local success, e = FTP.Remove(Dir)
		if success then
			self:send("250 Deleted FILENAME\n")
		else
			self:send("550 Could not delete FILENAME: "..e.."\n")
		end
		return true
	elseif string.sub(Line, 1, 4) == "STOR" then
		if self.Logged then
			local File = string.sub(Line, 6):gsub("/", "")

			if File:len() <= 0 then
				self:send("550 No file name\n")
				return true
			end

			local Dir, Path
			if self.Dir:len() > 0 then
				Dir = self.Dir.."/"..File
				Path = self.Dir
			else
				Dir = File
			end

			if not FTP.IsDir(Path) then
				if not Path then
					self:send("500 /"..Dir..": The system cannot find the path specified.\n")
					return true
				else
					local folders = {}
					for f in Path:gmatch("([^%/]+)") do
						table.insert(folders, f)
					end

					local NewPath = ""
					for i = 1, #folders do
						NewPath = NewPath.."/"..folders[i]
						if not FTP.IsDir(NewPath) then
							FTP.MkDir(NewPath)
							if not FTP.IsDir(NewPath) then
								self:send("500 /"..Dir..": Failed to create directory.\n")
								return true
							end
						end
					end
				end
			end

			if not self.SSTART then
				self.SSTART = socket.gettime()
			end

			if self:ConnectData() then
				if not self.SWARN then
					self.SFILE = FTP.Open(Dir, "wb")
					self.SWARN = true
					self:send("150 Accepted data connection\n")
				end

				local data, e = self.DataSocket:receive("*a")
				if data then
					if self.SFILE then
						self.SFILE:write(data)
					end
				elseif e == "closed" then
					self.DataSocket = nil
					if self.SFILE then
						self.SFILE:close()
						self.SFILE = nil
					end
					self.SWARN = nil
					self.SSTART = nil
					self:send("226 File successfully transferred\n")
					return true
				end
			elseif socket.gettime() - self.SSTART >= 2 then
				self.SSTART = nil
				self.SWARN = nil
				self:send("425 Can't open data connection\n")
				return true
			end
		else
			self:send("530 You aren't logged in\n")
			return true
		end
	elseif string.sub(Line, 1, 4) == "RETR" then
		if self.Logged then
			local File = string.sub(Line, 6):gsub("/", "")

			if #File <= 0 then
				self:send("550 No file name\n")
				return true
			end

			local Dir
			if #self.Dir > 0 then
				Dir = self.Dir.."/"..File
			else
				Dir = File
			end

			local FileObject = FTP.Open(Dir, "rb")
			if FileObject then
				FileObject:close()
			else
				self:send("550 Can't open "..File..": No such file or directory\n")
				return true
			end

			if not self.RSTART then
				self.RSTART = socket.gettime()
			end

			if self:ConnectData() then
				if not self.RWARN then
					self.RWARN = true
					self.RFILE = FTP.Open(Dir, "rb")
					self:send("150 Accepted data connection\n")
				end

				if self.RFILE then
					local data = self.RFILE:read(FTP.Configuration.TransferSpeed)
					if data then
						self.DataSocket:send(data)
					else
						self.RFILE:close()
						self.DataSocket:close()
						self.RFILE = nil
						self.RSTART = nil
						self.RWARN = nil
						self:send("226 File successfully transferred\n")
						return true
					end
				else
					self.RSTART = nil
					self.RWARN = nil
					self.DataSocket:close()
					self:send("226 File successfully transferred\n")
					return true
				end
			elseif socket.gettime() - self.RSTART > 2 then
				self.RSTART = nil
				self.RWARN = nil
				self:send("425 Can't open data connection\n")
				return true
			end
		else
			self:send("530 You aren't logged in\n")
			return true
		end
	else
		self:send("502 Command not implemented.\n")
		return true
	end
end

addhook("always", "FTP.Process")
function FTP.Process()
	FTP.UpdateServer()
	FTP.UpdateClients()
end

addhook("parse", "FTP.Console")
function FTP.Console(Command)
	if Command:sub(1, 3) == "ftp" then
		Command = Command:sub(5)
		if Command:sub(1, 7) == "adduser" then
			local Username, Password = Command:match("adduser (%w+) (%w+)")
			if not Username then
				FTP.Print("User name can only contain alphanumeric characters")
				return 1
			elseif not Password then
				FTP.Print("Password can only contain alphanumeric characters")
				return 1
			end
			if FTP.User[Username] then
				FTP.User[Username][1] = Password
				FTP.Save.Users()
				FTP.Print("Changed user password from '"..Username.."' to '"..string.rep("*", #Password).."'")
			else
				FTP.User[Username] = {Password, ""}
				FTP.Save.Users()
				FTP.Print("Created user '"..Username.."' with password '"..string.rep("*", #Password).."'")
			end
			return 1
		elseif Command:sub(1, 7) == "remuser" then
			local Username = Command:match("remuser (%w+)")
			if not Username then
				FTP.Print("User name can only contain alphanumeric characters")
				return 1
			end
			if FTP.User[Username] then
				FTP.User[Username] = nil
				FTP.Save.Users()
				FTP.Print("Removed user '"..Username.."'")
			else
				FTP.Print("User '"..Username.."' does not exist")
			end
			return 1
		elseif Command:sub(1, 8) == "userpath" then
			local Username, Path = Command:match("userpath (%w+) (.+)")
			if not Username then
				FTP.Print("User name can only contain alphanumeric characters")
				return 1
			elseif not Path then
				FTP.Print("Invalid path")
				return 1
			end

			if FTP.User[Username] then
				FTP.User[Username][2] = Path
				FTP.Save.Users()
				FTP.Print("Changed user path from '"..Username.."' to '"..Path.."'")
			else
				FTP.Print("User '"..Username.."' does not exist")
			end
			return 1
		elseif Command:sub(1, 4) == "port" then
			local Port = Command:match("port (%d+)")
			if not Port then
				FTP.Print("Invalid port")
				return 1
			end

			FTP.Configuration.Port = tonumber(Port)
			if FTP.Save then
				FTP.Save.Configuration()
			end
			return 1
		elseif Command:sub(1, 7) == "localip" then
			local IP = Command:match("localip ([%w|%p]+)")
			if not IP then
				FTP.Print("Invalid local ip")
				return 1
			end

			FTP.Configuration.IP = IP
			if FTP.Save then
				FTP.Save.Configuration()
			end
			return 1
		elseif Command:sub(1, 13) == "transferspeed" then
			local TransferSpeed = Command:match("transferspeed (%d+)")
			if not TransferSpeed then
				FTP.Print("Invalid transfer speed")
				return 1
			end
			FTP.Configuration.TransferSpeed = tonumber(TransferSpeed) or FTP.Configuration.TransferSpeed or 256
			if FTP.Save then
				FTP.Save.Configuration()
			end
			return 1
		elseif Command:sub(1, 5) == "start" then
			FTP.Start()
			FTP.Configuration.Start = true
			if FTP.Save then
				FTP.Save.Configuration()
			end
			return 1
		elseif Command:sub(1, 4) == "stop" then
			FTP.Stop()
			FTP.Configuration.Start = nil
			if FTP.Save then
				FTP.Save.Configuration()
			end
			return 1
		end
	end
end

function FTP.Load()
	FTP.Users = {}
	FTP.Configuration = {}

	local f = io.open(FTP.DIR.."ftp_config.cfg", "r")
	if f then
		for Command in f:lines() do
			parse(Command)
		end
		f:close()
	else
		FTP.Print("Failed to load settings")
	end

	local Users = io.open(FTP.DIR.."ftp_users.cfg", "r")
	if Users then
		local LineNumber = 0
		for Line in Users:lines() do
			LineNumber = LineNumber + 1

			local Username, Password, Path = Line:match("(%w+) (%w+) (.+)")
			if not Username then
				FTP.Print("Failed to load user name at line "..LineNumber)
			elseif not Password then
				FTP.Print("Failed to load user password at line "..LineNumber)
			elseif not Path then
				FTP.Print("Failed to load user path at line "..LineNumber)
			else
				FTP.Users[Username] = {Password, Path}
			end
		end
		Users:close()
	else
		FTP.Print("Failed to load users")
	end
	FTP.Load = nil
end; FTP.Load()

FTP.Save = {}
function FTP.Save.Users()
	local f = io.open(FTP.DIR.."ftp_users.cfg", "w")
	if f then
		local List = ""
		for Username, Memory in pairs(FTP.Users) do
			List = List .. Username .." " .. Memory[1] .. " ".. Memory[2] .. "\n"
		end
		f:write(List:sub(-2))
		f:close()
		FTP.Print("Saved users successfully")
	else
		FTP.Print("Failed to save users")
	end
end

function FTP.Save.Configuration()
	local f = io.open(FTP.DIR.."ftp_config.cfg", "w")
	if f then
		f:write("ftp port "..FTP.Configuration.Port.."\n")
		f:write("ftp localip "..FTP.Configuration.IP.."\n")
		f:write("ftp transferspeed "..FTP.Configuration.TransferSpeed.."\n")
		if FTP.Configuration.Start then
			f:write("ftp start\n")
		end
		f:close()
	end
end
